/* abs@well.com */

#include <stdio.h>
#include <avs/flow.h>
#include <avs/avs_data.h>
#include <avs/field.h>
#include <avs/geom.h>
#include <avs/colormap.h>
/* IAC CODE CHANGE : #include <avs/avs_math.h> */
#include <avs/avs_math.h>

/* To avoid obfuscating the code, put in a fixed polyhedron size */
#define MAX_PLIST_SIZE 1000000

AVSinit_modules()
{
    int AC_palms();
  
    AVSmodule_from_desc(AC_palms);
}

int AC_palms()
{
    int AC_palms_compute();

    int in_port1;
    int in_port2;

    /* Set the module name and type */
    AVSset_module_name("palms", MODULE_MAPPER);
  
    /* Input Port Specifications               */
    in_port1 = AVScreate_input_port("stream", "field 1D 2-space float", 
        REQUIRED);
  
    /* Create an output port for the result */
    AVScreate_output_port("Geometry", "geom");
  
    AVSadd_float_parameter("major_radius", 2.0, 0.00001, 1000.0);
    AVSadd_float_parameter("minor_radius", 1.0, 0.00001, 1000.0);
    AVSadd_parameter("major_facets","integer",4,3,100);
    AVSadd_parameter("minor_facets","integer",4,3,100);
    AVSadd_parameter("vector_length","integer",0,0,1000);
    AVSadd_float_parameter("angle_spread", 3.1415926, 0.0, 1000.0);
    AVSadd_float_parameter("z_offset", 0.0, 0.0, 1000.0);
    AVSadd_float_parameter("start_red", 1.0, 0.0, 1.0);
    AVSadd_float_parameter("start_green", 0.0, 0.0, 1.0);
    AVSadd_float_parameter("start_blue", 0.0, 0.0, 1.0);
    AVSadd_float_parameter("finish_red", 0.0, 0.0, 1.0);
    AVSadd_float_parameter("finish_green", 1.0, 0.0, 1.0);
    AVSadd_float_parameter("finish_blue", 0.0, 0.0, 1.0);
  
    /* Tell avs what subroutine to call to do the compute */
    AVSset_compute_proc(AC_palms_compute);
}

int AC_palms_compute(AVSfield_float *stream,
    GEOMedit_list *output,
    float *major_radius, float *minor_radius,
    int major_facets, int minor_facets, int vector_length,
    float *angle_spread, float *z_offset,
    float *start_red, float *start_green, float *start_blue,
    float *finish_red, float *finish_green, float *finish_blue)
{  
    GEOMobj *obj;
    float theta, max_theta, delta, old_theta;
    int index, flags, cindex;
    int v;
    float delta_angle;
    float r, g, b;
    float factor;

    float pi = 4.0*atan(1.0);
    float major_angle = 2.0*pi/(float)major_facets;
    float minor_angle = 2.0*pi/(float)minor_facets;

    int major_i, minor_i, facet_num, ppp, p1, p2, p3, p4;
    float major_a, minor_a;
    float vertex[MAX_PLIST_SIZE][3];
    float colors[MAX_PLIST_SIZE][3];
    int numverts;

    int plist[MAX_PLIST_SIZE];

    float x, y, z;
    float new_x, new_y, new_z;
    float _x_offset, _y_offset;

    double atan();
    double sin();
    double cos();

    char type[100];
    register int i, k;
    int n, pind;
    register int *vertex_data;
    int flag;
    FILE *f;
    int facet;
    int dims[99];
    float *fp1;
    int torus_num;
    float *fa1;

#ifdef DEBUG
    fprintf(stderr, "*major_radius = %f\n", *major_radius);
    fprintf(stderr, "*minor_radius = %f\n", *minor_radius);
    fprintf(stderr, "major_facets = %d\n", major_facets);
    fprintf(stderr, "minor_facets = %d\n", minor_facets);
    fprintf(stderr, "major_angle = %f\n", major_angle);
    fprintf(stderr, "minor_angle = %f\n", minor_angle);
    fprintf(stderr, "vector_length = %f\n", vector_length);
    fprintf(stderr, "*angle_spread = %f\n", *angle_spread);
    fprintf(stderr, "*z_offset = %f\n", *z_offset);
    fprintf(stderr, "*start_red = %f\n", *start_red);
    fprintf(stderr, "*finish_red = %f\n", *finish_red);
#endif DEBUG

    numverts = major_facets*minor_facets;

#ifdef DEBUG
    fprintf(stderr, "checkpoint #1\n");
#endif DEBUG

    AVSfield_get_dimensions(stream, dims);
    fa1 = AVSfield_points_ptr(stream);
    ppp = 0;
    delta_angle = *angle_spread/(float)vector_length;
    for (torus_num = 0; torus_num < dims[0]; torus_num++) {
        fp1 =  I1DV(stream, torus_num);

        for (v = 0; v <= vector_length; v++) {
            facet_num = 0;
            for (major_i = 0; major_i < major_facets; major_i++) {
                for (minor_i = 0; minor_i < minor_facets; minor_i++) {
                    major_a = major_angle*(float)major_i;
                    minor_a = minor_angle*(float)minor_i;
    
                    x = (*major_radius) + (*minor_radius)*(float)cos(minor_a);
                    y = 0.0;
                    z =                   (*minor_radius)*(float)sin(minor_a);
    
                    new_x = x*(float)cos(major_a);
                    new_y = x*(float)sin(major_a);
                    new_z = z;

                    x = new_x; y = new_y; z = new_z;
   
                    _x_offset = fa1[torus_num];
	            _y_offset = fa1[dims[0] + torus_num];

                    if (v > 0) {

                        /* scale torus by data amount */

                        new_x = 0.5*x;
                        new_y = 0.5*y;
                        new_z = 0.5*z;

                        x = new_x; y = new_y; z = new_z;

                        /* rotate torus about x axis 90 degrees */

                        new_x = x;
                        new_y =  y*(float)cos(pi/2.0) + z*(float)sin(pi/2.0);
                        new_z = -y*(float)sin(pi/2.0) + z*(float)cos(pi/2.0);

                        x = new_x; y = new_y; z = new_z;

                        /* translate torus along x axis */

                        new_x = x + *major_radius + 0.5*(*major_radius);
                        new_y = y;
                        new_z = z;

                        x = new_x; y = new_y; z = new_z;

                        /* rotate torus about z axis by correct angle */

                        new_x =  x*(float)cos((float)v*delta_angle)
                               + y*(float)sin((float)v*delta_angle);
                        new_y = -x*(float)sin((float)v*delta_angle)
                               + y*(float)cos((float)v*delta_angle);
                        new_z = z;

                        x = new_x; y = new_y; z = new_z;

	                factor = (float)(v - 1)/(float)(vector_length - 1);
		        r = factor*(*start_red)
                            + (1.0 - factor)*(*finish_red);
		        g = factor*(*start_green)
                            + (1.0 - factor)*(*finish_green);
		        b = factor*(*start_blue)
                            + (1.0 - factor)*(*finish_blue);
                        /*
                        r = 1.0; g = 1.0; b = 1.0;
                         */
                        }
                    else {
                        r = 0.7; g = 0.7; b = 0.7;
                        }

                    vertex[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][0] = x + _x_offset;
                    vertex[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][1] = y + _y_offset;
                    vertex[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][2] = z + *z_offset;
                    colors[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][0] = r;
                    colors[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][1] = g;
                    colors[facet_num + torus_num*numverts
                        + v*dims[0]*numverts][2] = b;
    
#ifdef DEBUG
    fprintf(stderr, "torus_num = %d\n", torus_num);
    fprintf(stderr, "facet_num = %d\n", facet_num);
#endif DEBUG
                    /* plist */
    
                    p1 = ++facet_num;
    
                    p2 = p1 + 1;
                    if (p2 > (major_i + 1)*minor_facets) {
                        p2 = p2 - minor_facets;
                        }

                    p4 = p1 + minor_facets;
                    if (p4 > major_facets*minor_facets) {
                        p4 = p4 - major_facets*minor_facets;
                        }
    
                    p3 = p4 + 1;
                    if (p3 > (major_i + 2)*minor_facets) {
                        p3 = p3 - minor_facets;
                        }
                    if (p1 == major_facets*minor_facets) {
                        p3 = p3 - minor_facets;
                        }
    
                    plist[ppp++] = 4;
                    plist[ppp++] = p1 + torus_num*numverts + v*dims[0]*numverts;
                    plist[ppp++] = p2 + torus_num*numverts + v*dims[0]*numverts;
                    plist[ppp++] = p3 + torus_num*numverts + v*dims[0]*numverts;
                    plist[ppp++] = p4 + torus_num*numverts + v*dims[0]*numverts;
                    }
                }
            }
        }
    plist[ppp] = 0;

    /* Create an empty polyhedron object */
    obj = GEOMcreate_obj(GEOM_POLYHEDRON,NULL);

    GEOMadd_vertices(obj,(float *)vertex,
        dims[0]*numverts*(vector_length + 1), GEOM_COPY_DATA);
    GEOMadd_float_colors(obj, (float *)colors,
        dims[0]*numverts*(vector_length + 1), GEOM_COPY_DATA);

    /* 
     * Add the polygons, copying the data since we didn't allocate it
     * dynamically.  If the concave option is passed, polygon can
     * contain concave polygons that will be handled properly when rendered
     */
    flag = 0;
    GEOMadd_polygons(obj,plist,flag,GEOM_COPY_DATA);

    /*
    facet = GEOM_FACET_NORMALS;
     */
    facet = 0;
    GEOMgen_normals(obj,facet);

/*******************/
    *output = GEOMinit_edit_list(*output);

    GEOMedit_geometry(*output,"palms",obj);

    GEOMdestroy_obj(obj);

    /* free the allocated space */

    return(1);
}

